# -*- coding: utf-8 -*-
"""
Helper functions to plot results.
"""
import re
import os.path as op

import nilearn.image      as niimg
import numpy              as np
import pandas             as pd
import scipy.io           as sio
from   nilearn.input_data import NiftiMasker
from   nilearn.image      import iter_img
from   nilearn.masking    import apply_mask
from   nilearn._utils.niimg_conversions import check_niimg, _index_img
from   boyle.nifti.utils  import filter_icc

from .utils import (get_largest_blobs,
                    build_raw_loadings_table,
                    add_groups_to_loadings_table,)
from ..interfaces import (plot_all_components,
                          plot_ica_components,
                          plot_multi_slices,
                          plot_overlays)
from ..utils import fetch_one_file


def plot_connectivity_matrix(connectivity_matrix, label_names):
    """ Plot the connectivity matrix. """
    # Display the correlation matrix
    import numpy as np
    from matplotlib import pyplot as plt

    plt.figure(figsize=(10, 10))

    # Mask out the major diagonal
    np.fill_diagonal(connectivity_matrix, 0)
    fig = plt.imshow(connectivity_matrix, interpolation="nearest", cmap="RdBu_r",
                     vmax=0.8, vmin=-0.8)
    plt.colorbar()
    # And display the labels
    _ = plt.xticks(range(len(label_names)), label_names, rotation=90)
    _ = plt.yticks(range(len(label_names)), label_names)

    return fig


def plot_ica_results(ica_result_dir, application, mask_file='', zscore=2., mode='+', bg_img=''):
    """  Use nilearn to plot results from different ICA analysis tools, given the ICA result folder path.

    Parameters
    ----------
    ica_result_dir: str
        Path to the ICA output folder or the ICA components volume file.

    application: str
        Choicese: ('nilearn', 'sbm', 'gift', 'gift-group')

    mask_file: str
        Path to the brain mask file to be used for thresholding.

    mode: str
        Choices: '+' for positive threshold,
                 '+-' for positive and negative threshold and
                 '-' for negative threshold.

    zscore: int
        Value of the Z-score thresholding.

    bg_img: str
        Path to a background image.
        If empty will use the SPM canonical brain image at 2mm.
    """
    if application == 'sbm': # SBM ICA (3D sources)
        plotter = SBMICAResultsPlotter(ica_result_dir)
    elif application == 'gift-group': # group GIFT ICA (4D sources)
        plotter = GIFTGroupICAResultsPlotter(ica_result_dir)
    elif application == 'gift': # single subject GIFT ICA (4D sources)
        plotter = GIFTICAResultsPlotter(ica_result_dir)
    elif application == 'canica': # group ICA (4D sources)
        plotter = CanICAResultsPlotter(ica_result_dir)
    else:
        raise NotImplementedError('Got no ICAResultsPlotter for application {}.'.format(application))

    plotter.fit(mask_file=mask_file, mode=mode, zscore=zscore)
    plotter.plot_icmaps(bg_img=bg_img)


class ICAResultsPlotter(object):
    """ Use nilearn to plot results from different ICA analysis tools, given the ICA result folder path.

    Parameters
    ----------
    ica_result_dir: str
        Path to the ICA output folder or the ICA components volume file.

    mask_file: str
        Path to the brain mask file to be used for thresholding.

    mode: str
        Choices: '+' for positive threshold,
                 '+-' for positive and negative threshold and
                 '-' for negative threshold.

    zscore: int or float
        Value of the Z-score thresholding.
    """
    _comps_fname = None

    def __init__(self, ica_result_dir):
        if not op.exists(ica_result_dir):
            raise IOError('Expected an existing file or folder, but could not find {}.'.format(ica_result_dir))

        self.ica_dir = op.expanduser(ica_result_dir)

    def fit(self,  mask_file='', mode='+-', zscore=2):
        """ Process/filter/threshold the output to make it ready for plot.

        Parameters
        ----------
        mask_file: str
            Path to the brain mask file to be used for thresholding.

        mode: str
            Choices: '+' for positive threshold,
                     '+-' or '-+' for positive and negative threshold and
                     '-' for negative threshold.

        zscore: int or float
            Value of the Z-score thresholding.
        """
        self.mask_file = mask_file
        self.mode = mode
        self.zscore = zscore
        self._icc_imgs = None
        self._update(force=True)

    def _update(self, force=False):
        if self._icc_imgs is not None and not force:
            return

        self._check_output()
        self._icc_imgs = self._filter_ic_imgs(self._fetch_components_file())

    def is_fit(self):
        """ Return True if the self `fit` function has been called, False otherwise."""
        if not hasattr(self, '_icc_imgs'):
            return False
        else:
            return self._icc_imgs is not None

    def _check_is_fit(self):
        if not self.is_fit():
            raise RuntimeError('You should call `fit()` before using this function.')

    def _fetch_components_file(self):
        if self._comps_fname is None:
            raise NotImplementedError('This is a generic class to support the output from different applications,'
                                      'please use a derived class from ICAResultsPlotter.')

        return fetch_one_file(self.ica_dir, self._comps_fname)

    def _filter_ic_imgs(self, ic_file):
        if self.zscore > 0:
            do_zscore = True
        else:
            do_zscore = False

        mask = niimg.load_img(self.mask_file)
        return [filter_icc(icimg, mask=mask, thr=self.zscore, zscore=do_zscore, mode=self.mode)
                for icimg in iter_img(ic_file)]

    def _check_output(self):
        raise NotImplementedError('This is a generic class to support the output from different applications,'
                                  'please use a derived class from ICAResultsPlotter.')

    def plot_icmaps(self, outtype='png', **kwargs):
        """ Plot the thresholded IC spatial maps and store the outputs in the ICA results folder.
        Parameters
        ----------
        outtype: str
            Extension (without the '.') of the output files, will specify which plot image file you want.

        Returns
        -------
        all_icc_plot_f: str

        iccs_plot_f: str

        sliced_ic_plots: list of str
        """
        # specify the file paths
        all_icc_plot_f  = op.join(self.ica_dir, 'all_components_zscore_{}.{}'.format(self.zscore, outtype))
        iccs_plot_f     = op.join(self.ica_dir,  'ic_components_zscore_{}.{}'.format(self.zscore, outtype))
        icc_multi_slice = op.join(self.ica_dir, 'ic_map_{}_zscore_{}.{}')

        # make the plots
        fig1 = plot_ica_components(self._icc_imgs, **kwargs)
        fig1.savefig(iccs_plot_f, facecolor=fig1.get_facecolor(), edgecolor='none')

        fig2 = plot_all_components(self._icc_imgs, **kwargs)
        fig2.savefig(all_icc_plot_f, facecolor=fig2.get_facecolor(), edgecolor='none')

        # make the multi sliced IC plots
        sliced_ic_plots = []
        for i, img in enumerate(iter_img(self._icc_imgs)):
            fig3 = plot_multi_slices(img,
                                     cut_dir="z",
                                     n_cuts=24,
                                     n_cols=4,
                                     title="IC {}\n(z-score {})".format(i+1, self.zscore),
                                     title_fontsize=32,
                                     plot_func=None,
                                     **kwargs)

            # prepare the output file name/path
            out_f = icc_multi_slice.format(i+1, self.zscore, outtype)
            fig3.savefig(out_f, facecolor=fig3.get_facecolor(), edgecolor='none')
            sliced_ic_plots.append(out_f)

        return all_icc_plot_f, iccs_plot_f, sliced_ic_plots


class CanICAResultsPlotter(ICAResultsPlotter):
    """ Use nilearn to plot results from CanICA and DictLearing tools from nilearn, given the ICA result folder path.

    Parameters
    ----------
    ica_result_dir: str
        Path to the ICA output folder or the ICA components volume file.
    """
    _comps_fname = 'canica_resting_state.nii.gz'


class MIALABICAResultsPlotter(ICAResultsPlotter):
    """ Use nilearn to plot results from MIALAB (GIFT and SBM) tools, given the ICA result folder path.

    Parameters
    ----------
    ica_result_dir: str
        Path to the ICA output folder or the ICA components volume file.
    """
    # define the input file patterns
    _tcs_fname       = '*_timecourses_ica*'
    _comps_fname     = '*_component_ica*'
    _subjects_fname  = '*Subject.mat'
    _mask_fname      = r'.*Mask.(hdr|nii)'

    def __init__(self, ica_result_dir):
        super(MIALABICAResultsPlotter, self).__init__(ica_result_dir=ica_result_dir)
        self._subjid_pat = ''

    def _parse_groups_file(self, group_labels_file):
        """ Return the list of groups and subject ids from the group_labels_file and the Subjects.mat file.
        the file paths inside the *Subject.mat file.
        One file is `subject_loadings.xls` which has the loadings as is, with the subjects IDs and group.
        The other file is `subject_group_loadings.xls` which has the loading signs changed according to
        the average correlation value of the "main" region of each of the IC spatial maps.

        Parameters
        ----------
        group_labels_file: str
            A CSV file with two columns: "subject_id" and "group".
            The subject_ids must be in the paths contained in the Subject.mat
            file and match the `subjid_pat` argument.

        Returns
        -------
        groups: pandas.DataFrame
        """
        groups = None
        if group_labels_file and op.exists(group_labels_file):
            groups = pd.read_csv(group_labels_file)
            if not 'subject_id' in groups.columns or \
               not 'group' in groups.columns:
                raise AttributeError("Please add columns names 'subject_id' "
                                     "and 'group' to the group labels file.")
        return groups

    def _load_components(self):
        """ Return components image file and check shape match."""
        compsf = self._fetch_components_file()
        comps_img = niimg.load_img(compsf)
        return comps_img

    def _load_loadings(self):
        loadf = fetch_one_file(self.ica_dir, self._tcs_fname)
        loads = niimg.load_img(loadf).get_data()
        return loads

    def _fetch_components_file(self):
        return fetch_one_file(self.ica_dir, self._comps_fname)

    def _get_subject_files(self):
        """ Load the .mat file with subjects lists, mainly to get the order in
        which subjects are introduced in the other matrices.
        """
        from itertools import chain

        subjsf   = fetch_one_file(self.ica_dir, self._subjects_fname)
        mat_file = sio.loadmat(subjsf)['files']
        return [f.strip() for f in list(chain.from_iterable(chain.from_iterable(chain.from_iterable(mat_file))))]

    def _get_subject_ids(self, subjid_pat):
        """ Return the list of subject ids parsed from the file paths present in the Subjects.mat file.

        Parameters
        ----------
        subjid_pat: regext str
            A search regex pattern that returns one group element that
            contains the subject id.
            This will be used to search for subject_id in the file paths
            contained in the Subjects.mat file.

        Returns
        -------
        patids: list[str]
        """
        self._subjid_pat = subjid_pat

        subj_files = self._get_subject_files()

        # process the paths to extract the patient IDs given by `patid_re`
        in_files = [f.split(',')[0] for f in subj_files]
        patid_re = re.compile(subjid_pat, re.I)
        patids   = [patid_re.search(f).group() for f in in_files]

        return patids

    @staticmethod
    def _index_img(img_file, index):
        """ Return the volume in `index` 4th dimension index of `img_file`.
        If the image is 3D and idx is zero will return the container of `img_file`.
        """
        imgs = check_niimg(img_file, ensure_ndim=4, atleast_4d=True)
        return _index_img(imgs, index)

    def load_mask(self):
        """ Return the mask image. """
        mask_file = fetch_one_file(self.ica_dir, self._mask_fname, pat_type='re.match')
        return niimg.load_img(mask_file)

    def load_subject_data(self, masked=False, **kwargs):
        """ Generator of the input data volumes in the order it was inserted in GIFT ICA.

        Returns
        -------
        subj_imgs: generator of niimg-like objects.
            The input data to the ICA.

        masked: bool
            If True will apply the mask to the data.

        kwargs: keyword arguments
            Keyword arguments for the NiftiMasker used to mask the data.
            The mask used is the one used in the `fit` function or specified
            in `_mask_fname`.

        Note
        ----
        This will actually read the paths in the Subjects.mat file. If you have moved
        or erased those files, this will not work as expected.
        """
        if masked:
            sessions = self._input_data_sessions()
            data = self._apply_mask_to_imgs(self._load_subject_data(), sessions=sessions, **kwargs)
        else:
            data = self._load_subject_data()

        return np.array(list(data))

    def load_components_data(self, masked=False, **kwargs):
        if masked:
            comps = self._load_components()
            n_comps = comps.shape[-1]
            return self._apply_mask_to_4dimg(comps, **kwargs)
        else:
            return self._load_components()

    def _input_data_sessions(self):
        """ Return a 1D array with a session number for each session
        in the input data.
        This is used for NiftiMasker as a value for the `sessions`
        parameter.

        Example
        -------
        Given input data:
        ["subject01.nii,1",
         "subject01.nii,2",
         "subject02.nii,1",]

        Would return: [0, 0, 1]
        """
        # get the subject file list
        subj_files = self._get_subject_files()

        # remove the numbers after the comma in this list
        files = np.array([f.split(',')[0] for f in subj_files])

        # make a dict that assign a number to each unique file
        file_int = {f: i for i, f in enumerate(np.unique(files))}

        # return the corresponding numbers for each file in the cleaned file list
        return np.array([file_int[f] for f in files])

    def _load_subject_data(self):
        img_file_list = self._get_subject_files()
        for line in img_file_list:
            img_file, idx = line.split(',')
            yield self._index_img(img_file, int(idx) - 1)

    def _apply_mask_to_img(self, img, **kwargs):
        masker = NiftiMasker(mask_img=self.load_mask(), **kwargs)
        return masker.fit_transform(img)

    def _apply_mask_to_4dimg(self, imgs, **kwargs):
        masker = NiftiMasker(mask_img=self.load_mask(), **kwargs)
        return (masker.fit_transform(img) for img in iter_img(imgs))

    def _apply_mask_to_imgs(self, imgs, **kwargs):
        masker = NiftiMasker(mask_img=self.load_mask(), **kwargs)
        return (masker.fit_transform(img) for img in imgs)

    def fit(self,  mask_file='', mode='+', zscore=2):
        """ Process/filter/threshold the output to make it ready for plot.
        If no mask_file is set, will pick the output from GIFT.

        Parameters
        ----------
        mask_file: str, optional
            Path to the brain mask file to be used for thresholding.

        mode: str, optional
            Choices: '+' for positive threshold,
                     '+-' for positive and negative threshold and
                     '-' for negative threshold.

        zscore: int or float, optional
            Value of the Z-score thresholding.
        """
        if not mask_file:
            mask_file = fetch_one_file(self.ica_dir, self._mask_fname, pat_type='re.match')

        super(MIALABICAResultsPlotter, self).fit(mask_file=mask_file,
                                                 mode=mode,
                                                 zscore=zscore)

    def simple_loadings_df(self, group_labels_file, subjid_pat=r'(?P<patid>[a-z]{2}_[0-9]{6})'):
        """ Return a pandas.DataFrame spreadsheet ready for an excel file with the subject IDs taken from
        the file paths inside the *Subject.mat file.
        One file is `subject_loadings.xls` which has the loadings as is, with the subjects IDs and group.
        The other file is `subject_group_loadings.xls` which has the loading signs changed according to
        the average correlation value of the "main" region of each of the IC spatial maps.

        Parameters
        ----------
        group_labels_file: str
            A CSV file with two columns: "subject_id" and "group".
            The subject_ids must be in the paths contained in the Subject.mat
            file and match the `subjid_pat` argument.

        subjid_pat: regext str
            A search regex pattern that returns one group element that
            contains the subject id.
            This will be used to search for subject_id in the file paths
            contained in the Subjects.mat file.

        Returns
        -------
        loadings_df: pandas.DataFrame
        """
        # make sure file exists
        if not op.exists(group_labels_file):
            raise FileNotFoundError('The file {} has not been found.'.format(group_labels_file))

        # make sure this object has been .fit()
        self._update()

        # read the groups file
        groups = self._parse_groups_file(group_labels_file=group_labels_file)
        patids = self._get_subject_ids(subjid_pat=subjid_pat)

        # load the loadings file
        loads = self._load_loadings()

        # build the raw loadings table
        df = build_raw_loadings_table(loads, patids)
        df = add_groups_to_loadings_table(df, groups)
        #df.to_excel(op.join(ica_out_dir, rawloadings_filename))

        return df

    def weighted_loadings_df(self, group_labels_file, subjid_pat=r'(?P<patid>[a-z]{2}_[0-9]{6})'):
        """ Return a pandas.DataFrame with the loadings, subject IDs and groups.
        This version of the sheet has the loadings signs changed for each group of subjects.
        Multiply the largest blob of each IC spatial map (see `plot_ic_largest_blobs`) and the sign of the
        average loading of each group.

        Parameters
        ----------
        group_labels_file: str
            A CSV file with two columns: "subject_id" and "group".
            The subject_ids must be in the paths contained in the Subject.mat
            file and match the `subjid_pat` argument.

        subjid_pat: regext str
            A search regex pattern that returns one group element that
            contains the subject id.
            This will be used to search for subject_id in the file paths
            contained in the Subjects.mat file.

        Returns
        -------
        loadings_df: pandas.DataFrame
        """
        # make sure file exists
        if not op.exists(group_labels_file):
            raise FileNotFoundError('The file {} has not been found.'.format(group_labels_file))

        # make sure this object has been .fit()
        self._update()

        # let's first pick the simple version of the loadings
        df = self.simple_loadings_df(group_labels_file, subjid_pat=subjid_pat)

        # get the values of the largest blobs in the filtered IC maps
        blobs = get_largest_blobs(self._icc_imgs)

        # apply the blob mask to each corresponding IC map
        masks = [apply_mask(ic_map, blob) for ic_map, blob in zip(self._icc_imgs, blobs)]

        # calculate the avg per blob
        blob_avgs = [mask.mean() for mask in masks]

        # multiply the avg blob value with the group loadings
        blob_signs = np.sign(blob_avgs)
        n_ics      = len(blob_avgs)

        df[list(range(1, n_ics+1))] = df[list(range(1, n_ics+1))] * blob_signs

        # put the dataframe back into one whole ungrouped dataframe
        return df

    def plot_icmaps_and_blobs(self, outfile, bg_img=None, **kwargs):
        """ Plot the IC maps with the largest blobs overlaid.
        Parameters
        ----------
        outfile: str
            Path to output plot file.

        bg_img: str
            Path to image file.
        """
        # make sure this object has been .fit()
        self._update()
        fig = plot_overlays(self._icc_imgs, get_largest_blobs(self._icc_imgs),
                            bg_img=bg_img, figsize=(2.5, 3), **kwargs)

        # save fig
        fig.savefig(outfile, facecolor=fig.get_facecolor(), edgecolor='none')


class GIFTICAResultsPlotter(MIALABICAResultsPlotter):
    """ Use nilearn to plot results from the MIALAB ICA tool used to analyse one subject,
    given the ICA result folder path.

    Parameters
    ----------
    ica_result_dir: str
        Path to the ICA output folder or the ICA components volume file.
    """
    _tcs_fname   = r'.*_sub[0]+1_timecourses_ica_s1_.nii'
    _comps_fname = r'.*_sub[0]+1_component_ica_s1_.nii'

    def _check_output(self):
        comps_img = self._load_components()
        tcs_img   = self._load_timecourses()

        n_ics = tcs_img.shape[1]
        if comps_img.get_data().shape[-1] != n_ics:
            raise AttributeError('Shape mismatch between timecourses matrix {} '
                                 'and components data shape {}.'.format(tcs_img.shape,
                                                                        comps_img.get_data().shape))

        # read the groups file
        subj_files   = self._get_subject_files()
        n_timepoints = len(subj_files)

        # check shapes
        if n_timepoints != tcs_img.shape[0]:
            raise AttributeError('Shape mismatch between list of subjects {} '
                                 'and timecourses data shape {}.'.format(n_timepoints,
                                                                         tcs_img.shape))

    def _load_timecourses(self):
        """ Return the timecourses image file and checks if the shape is correct."""
        # load the timecourses file
        tcsf = fetch_one_file(self.ica_dir, self._tcs_fname, pat_type='re.match')
        tcs = niimg.load_img(tcsf).get_data()
        return tcs

    def _fetch_components_file(self):
        return fetch_one_file(self.ica_dir, self._comps_fname, pat_type='re.match')


class GIFTGroupICAResultsPlotter(MIALABICAResultsPlotter):
    """ Use nilearn to plot results from the MIALAB Group ICA tool used to analyse a group of subjects,
    given the ICA result folder path.

    Parameters
    ----------
    ica_result_dir: str
        Path to the ICA output folder or the ICA components volume file.
    """
    _tcs_fname   = '*_agg__timecourses_ica_.nii'
    _comps_fname = '*_agg__component_ica_.nii'

    def _check_output(self):
        comps_img = self._load_components()
        tcs_img   = self._load_timecourses()

        n_ics = tcs_img.shape[1]
        if comps_img.get_data().shape[-1] != n_ics:
            raise AttributeError('Shape mismatch between timecourses matrix {} '
                                 'and components data shape {}.'.format(tcs_img.shape,
                                                                        comps_img.get_data().shape))

        # read the groups file
        # subj_files   = self._get_subject_files()
        # n_timepoints = len(subj_files)

        # check shapes
        #if n_timepoints != tcs_img.shape[0]:
        #    raise AttributeError('Shape mismatch between list of subjects {} '
        #                         'and timecourses data shape {}.'.format(n_timepoints,
        #                                                                 tcs_img.shape))

    def _load_timecourses(self):
        """ Return the timecourses image file and checks if the shape is correct."""
        # load the timecourses file
        tcsf = fetch_one_file(self.ica_dir, self._tcs_fname)
        tcs = niimg.load_img(tcsf).get_data()
        return tcs

    def _fetch_components_file(self):
        return fetch_one_file(self.ica_dir, self._comps_fname)

    # def _calculate_goodness_of_fit(self):
    #     """ Return the goodness-of-fit values from a GIFT result."""
    #     pass
    #     #TODO
    #
    # def goodness_of_fit_df(self, group_labels_file, subjid_pat=r'(?P<patid>[a-z]{2}_[0-9]{6})'):
    #     """ Return a pandas.DataFrame ready for an excel file with:
    #     - the subject IDs taken from the file paths inside the *Subject.mat file,
    #     - the groups taken from `group_labels_file`, and
    #     - the goodness-of-fit measures for each subject and independent component.
    #
    #     Parameters
    #     ----------
    #     group_labels_file: str
    #         A CSV file with two columns: "subject_id" and "group".
    #         The subject_ids must be in the paths contained in the Subject.mat
    #         file and match the `subjid_pat` argument.
    #
    #     subjid_pat: regext str
    #         A search regex pattern that returns one group element that
    #         contains the subject id.
    #         This will be used to *search* for subject_id in the file paths
    #         contained in the Subjects.mat file.
    #
    #     Returns
    #     -------
    #     gof_df: pandas.DataFrame
    #     """
    #     # make sure file exists
    #     if not op.exists(group_labels_file):
    #         raise FileNotFoundError('The file {} has not been found.'.format(group_labels_file))
    #
    #     # make sure this object has been .fit()
    #     self._update()
    #
    #     # read the groups file
    #     groups = self._parse_groups_file(group_labels_file=group_labels_file)
    #     patids = self._get_subject_ids(subjid_pat=subjid_pat)
    #
    #     # calculate the goodness of fit
    #     gofs = self._calculate_goodness_of_fit(patids)
    #
    #     # build the goodness-of-fit table
    #     df = build_raw_loadings_table(gofs, patids)
    #     df = add_groups_to_loadings_table(df, groups)
    #
    #     return df


class SBMICAResultsPlotter(MIALABICAResultsPlotter):
    """ Use nilearn to plot results from the MIALAB SBM tool, given the ICA result folder path.

    Parameters
    ----------
    ica_result_dir: str
        Path to the ICA output folder or the ICA components volume file.
    """
    _tcs_fname  = '*_loading_coeff_.nii'

    def _check_output(self):
        # read the groups file
        subj_files   = self._get_subject_files()
        n_timepoints = len(subj_files)

        # load the loadings file
        loads = self._load_loadings()

        # check shapes
        if n_timepoints != loads.shape[0]:
            raise AttributeError('Shape mismatch between list of subjects {} '
                                 'and loadings data shape {}.'.format(n_timepoints,
                                                                      loads.shape))

        # load components image file to check shape match
        # this doesn't make much sense to be here (because it does not use the components image
        # but it checks if the ICA output is consistent).
        compsf = fetch_one_file(self.ica_dir, self._comps_fname)
        comps_img = niimg.load_img(compsf)

        n_ics = loads.shape[1]
        if comps_img.get_data().shape[-1] != n_ics:
            raise AttributeError('Shape mismatch between loadings matrix {} '
                                 'and components data shape {}.'.format(loads.shape,
                                                                        comps_img.get_data().shape))


def ica_loadings_sheet(plotter, labels_file, mask=None, bg_img=None, zscore=2.,
                       subjid_pat=r'(?P<patid>[a-z]{2}_[0-9]{6})'):
    """ Save the Excel loadings files in the `ica_out_dir`.
    One file is `subject_loadings.xls` which has the loadings as is, with the subjects IDs and group.
    The other file is `subject_group_loadings.xls` which has the loading signs changed according to
    the average correlation value of the "main" region of each of the IC spatial maps.

    Parameters
    ----------
    plotter: MIALABICAResultsPlotter, or any derivative
        The GIFT ICA analysis plotter object.
        You must call the plotter fit function before, otherwise this function will call fit
        with `mask`, `zscore` and mode='+-'.

    labels_file: str
        A CSV file with two columns: "subject_id" and "group".
        The subject_ids must be in the paths contained in the Subject.mat
        file and match the `subjid_pat` argument.

    mask: str
        Path to a mask file to select only brain area from the IC spatial map.

    bg_img: str
        A background image for the blob plots check report, to verify that the blobs
        taken into account for the loadings signs are correct.

    zscore: float
        Value to threshold the IC spatial maps to obtain the IC spatial map "main" region.

    subjid_pat: regext str
        A search regex pattern that returns one group element that
        contains the subject id.
        This will be used to search for subject_id in the file paths
        contained in the Subjects.mat file.
    """
    rawloadings_filename   = 'subject_loadings.xls'
    grouploadings_filename = 'subject_weighted_loadings.xls'
    check_blob_plot        = 'check_sign_blobs.png'

    if not plotter.is_fit():
        plotter.fit(mask_file=mask, mode='+-', zscore=zscore)

    output_dir = plotter.ica_dir

    # generate and save the simple loadings sheet
    sdf = plotter.simple_loadings_df(group_labels_file=labels_file, subjid_pat=subjid_pat)
    sdf.to_excel(op.join(output_dir, rawloadings_filename))

    # generate and save the group-processed loadings sheet
    pdf = plotter.weighted_loadings_df(group_labels_file=labels_file, subjid_pat=subjid_pat)
    pdf.to_excel(op.join(output_dir, grouploadings_filename))

    # plot blobs over IC maps for checking
    check_blob_plot = op.join(output_dir, check_blob_plot)
    plotter.plot_icmaps_and_blobs(check_blob_plot, bg_img=bg_img)


# -*- coding: utf-8 -*-
"""
Functions to create pipelines for public and not so public available datasets.
"""

from collections import OrderedDict

from   .config  import update_config
from   .anat    import (attach_spm_anat_preprocessing,
                        attach_ants_cortical_thickness)
from   .dmri    import (attach_spm_fsl_dti_preprocessing,
                        attach_camino_tractography)
from   .fmri    import attach_rest_preprocessing, attach_rest_grptemplate_preprocessing
from   .io      import build_crumb_workflow
from   .pet     import (attach_spm_mrpet_preprocessing,
                        attach_spm_pet_preprocessing,
                        attach_spm_pet_grouptemplate,)


def _cobre_wf_setup(wf_name):
    """ Return a list of workflow-attach functions and a dict with kwargs for
    the `in_out_crumb_wf` function to run workflows against the COBRE database.

    Parameters
    ----------
    wf_name

    Returns
    -------
    wf_attachers: dict with wf attacher functions

    wf_params: dict with configuration parameters for the workflows.
        Use this only to fixed configurations needed for the correct functioning of the workflows.

    files_crumb_args: dict with kwargs
        This arguments are for the in_out_crumb_wf.
        A dict that declares the values each file type the crumb arguments in `data_crumb` must be replaced with.
        Example:
              {'anat':  [('modality', 'anat_1'),
                         ('image',    'mprage.nii.gz')],
              'rest':   [('modality', 'rest_1'),
                         ('image',    'rest.nii.gz')],
              }
    """
    attach_functions = {"spm_anat_preproc":         [("spm_anat_preproc",  attach_spm_anat_preprocessing)],

                        "spm_anat_rest_preproc":    [("spm_anat_preproc",  attach_spm_anat_preprocessing),
                                                     ("spm_rest_preproc",  attach_rest_preprocessing),
                                                    ],
                       }

    if wf_name not in attach_functions:
        raise KeyError('Expected an existing pipeline name, got {}. '
                       'Available options: {}'.format(wf_name, list(attach_functions.keys())))

    files_crumb_args = {'anat':  [('modality', 'anat_1'),
                                  ('image',    'mprage.nii.gz')]} #'anat_1/mprage.nii.gz',

    if 'rest' in wf_name:
        files_crumb_args.update({'rest':  [('modality', 'rest_1'),
                                           ('image',    'rest.nii.gz')], # 'rest_1/rest.nii.gz'},
                                })

    params = {'file_templates': files_crumb_args}

    return OrderedDict(attach_functions[wf_name]), params


def _clinical_wf_setup(wf_name):
    """ Return a list of workflow-attach functions and a dict with kwargs for
    the `in_out_crumb_wf` function to run workflows against the clinical database.

    Parameters
    ----------
    wf_name

    Returns
    -------
    wf_attachers: dict with wf attacher functions

    wf_params: dict with configuration parameters for the workflows.
        Use this only to fixed configurations needed for the correct functioning of the workflows.

    files_crumb_args: dict with kwargs
        This arguments are for the in_out_crumb_wf.
        A dict that declares the values each file type the crumb arguments in `data_crumb` must be replaced with.
        Example:
              {'anat':  [('modality', 'anat_1'),
                         ('image',    'mprage.nii.gz')],
              'rest':   [('modality', 'rest_1'),
                         ('image',    'rest.nii.gz')],
              }
    """
                        # MPRAGE bias-field correction, normalization to MNI, and tissue segmentation
    attach_functions = {"spm_anat_preproc":     [("spm_anat_preproc", attach_spm_anat_preprocessing)],

                        # PET normalization to MNI
                        "spm_pet_preproc":      [("spm_pet_preproc", attach_spm_pet_preprocessing)],

                        # will create a PET group template
                        "spm_pet_template":     [("spm_pet_preproc",       attach_spm_pet_preprocessing),
                                                 ("spm_pet_grouptemplate", attach_spm_pet_grouptemplate),
                                                ],

                        # MPRAGE preprocessing, PET MNI group template, PET PVC, and PET normalization to group template
                        "spm_anat_pet_tpm_pvc": [("spm_anat_preproc",        attach_spm_anat_preprocessing),
                                                 ("spm_pet_preproc",         attach_spm_pet_preprocessing),
                                                 ("spm_mrpet_grouptemplate", attach_spm_pet_grouptemplate),
                                                ],

                        # MPRAGE preprocessing, PET PVC, and PET normalization to MNI
                        "spm_anat_pet_pvc":     [("spm_anat_preproc",  attach_spm_anat_preprocessing),
                                                 ("spm_mrpet_preproc", attach_spm_mrpet_preprocessing),
                                                ],

                        # MPRAGE preprocessing, rs-fMRI preprocessing and normalization to MNI
                        "spm_anat_rest_preproc":[("spm_anat_preproc",  attach_spm_anat_preprocessing),
                                                 ("spm_rest_preproc",  attach_rest_preprocessing),
                                                ],

                        # MPRAGE preprocessing, DTI preprocessing with FSL
                        "fsl_dti_preproc":      [("spm_anat_preproc",  attach_spm_anat_preprocessing),
                                                 ("fsl_dti_preproc",   attach_spm_fsl_dti_preprocessing),
                                                ],

                        # MPRAGE preprocessing, DTI preprocessing with FSL, and tractography with Camino
                        "anat_dti_camino":      [("spm_anat_preproc",  attach_spm_anat_preprocessing),
                                                 ("fsl_dti_preproc",   attach_spm_fsl_dti_preprocessing),
                                                 ("camino_tract" ,     attach_camino_tractography),
                                                ],

                        # MPRAGE and PET preprocessing, DTI preprocessing with FSL, and tractography with Camino
                        "anat_pet_dti_camino":  [("spm_anat_preproc",  attach_spm_anat_preprocessing),
                                                 ("spm_mrpet_preproc", attach_spm_mrpet_preprocessing),
                                                 ("fsl_dti_preproc",   attach_spm_fsl_dti_preprocessing),
                                                 ("camino_tract",      attach_camino_tractography),
                                                ],

                        # MPRAGE preprocessing, and EPI group template
                        "spm_anat_rest_tpm_preproc": [("spm_anat_preproc", attach_spm_anat_preprocessing),
                                                      ("spm_warp_fmri",    attach_rest_grptemplate_preprocessing),
                                                     ],

                        # MPRAGE preprocessing, EPI and PET group template, and PET and rs-fMRI preprocessing and
                        # normalization to group template
                        "spm_anat_pet_rest_tpm_preproc": [("spm_anat_preproc",        attach_spm_anat_preprocessing),
                                                          ("spm_warp_fmri",           attach_rest_grptemplate_preprocessing),
                                                          ("spm_pet_preproc",         attach_spm_pet_preprocessing),
                                                          ("spm_mrpet_grouptemplate", attach_spm_pet_grouptemplate),
                                                         ],

                        # MPRAGE preprocessing, EPI group template, and rs-fMRI preprocessing and normalization to MNI
                        "spm_anat_pet_rest_preproc":    [("spm_anat_preproc",  attach_spm_anat_preprocessing),
                                                         ("spm_rest_preproc",  attach_rest_preprocessing),
                                                         ("spm_mrpet_preproc", attach_spm_mrpet_preprocessing),
                                                        ],

                        # MPRAGE and cortical thickness
                        "spm_anat_ants_cortical_thick": [("spm_anat_preproc",  attach_spm_anat_preprocessing),
                                                         ("ants_cort_thick",   attach_ants_cortical_thickness),
                                                        ],
                       }

    parameters       = {"spm_pet_template":     [('spm_pet_template.do_petpvc', False),],
                        "spm_anat_pet_tpm_pvc": [('spm_pet_template.do_petpvc', True),],
                       }

    if wf_name not in attach_functions:
        raise KeyError('Expected an existing pipeline name, got {}. '
                       'Available options: {}'.format(wf_name, list(attach_functions.keys())))

    # the pipeline parameters
    wf_params = parameters.get(wf_name, None)
    if wf_params is not None:
        wf_params = dict(wf_params)

    # the input files crumb patterns
    files_crumb_args = {}
    files_crumb_args.update({'anat':  [('image', 'anat_hc.nii.gz')]})

    if 'pet' in wf_name:
        files_crumb_args.update({'pet':  [('image', 'pet_fdg.nii.gz')],})

    if 'rest' in wf_name:
        files_crumb_args.update({'rest':  [('image', 'rest.nii.gz')],})

    if 'dti' in wf_name:
        files_crumb_args.update({'diff': [('image', 'diff.nii.gz')],
                                 'bval': [('image', 'diff.bval')],
                                 'bvec': [('image', 'diff.bvec')],
                                })

    return OrderedDict(attach_functions[wf_name]), wf_params, files_crumb_args


def cobre_crumb_workflow(wf_name, data_crumb, output_dir, cache_dir='', config_file='', params=None):
    """ Returns a workflow for the COBRE database.

    Parameters
    ----------
    wf_name: str
        A name for the workflow.
        Choices: 'spm_anat_preproc': MPRAGE preprocessing with SPM12
                 'spm_rest_preproc': MPRAGE+rs-fMRI preprocessing with SPM12

    data_crumb: hansel.Crumb
        The crumb until the subject files.
        Example: Crumb('/home/hansel/cobre/raw/{subject_id}/session_1/{modality}/{image})
        The last 2 crumb arguments of `data_crumb` must be '{modality}/{image}', which indicates each of the
        subject/session files. This argument will be replaced by the corresponding image name.

    cache_dir: str
        The working directory of the workflow.

    output_dir: str
        The output folder path

    config_file: str
        Path to a configuration file. Will use anything compatible with Kaptan.
        Have a look at config.py.

    params: dict with arguments
        crumb_replaces

        atlas_file
    """
    attach_funcs, cfg_params, file_templates = _cobre_wf_setup(wf_name)

    if params is None:
        params = {}

    if config_file:
        update_config(config_file)

    if params:
        update_config(params)

    if cfg_params is not None:
        update_config(cfg_params)

    wf = build_crumb_workflow(attach_funcs,
                              data_crumb=data_crumb,
                              in_out_kwargs=file_templates,
                              output_dir=output_dir,
                              cache_dir=cache_dir,)

    return wf


def clinical_crumb_workflow(wf_name, data_crumb, output_dir, cache_dir='', config_file='', params=None):
    """ Returns a workflow for the a clinical database.

    Parameters
    ----------
    wf_name: str
        A name for the workflow to be created.
        Choices: 'spm_anat_preproc':     MPRAGE preprocessing with SPM12
                 'spm_anat_pet_preproc': MPRAGE+FDG-PET preprocessing with SPM12
                 'spm_pet_preproc':      FDG-PET only preprocessing with SPM12

    data_crumb: hansel.Crumb
        The crumb until the subject files.
        Example: Crumb('/home/hansel/data/{subject_id}/{session_id}/{modality}/{image})
        The last crumb argument of `data_crumb` must be '{image}', which indicates each of the
        subject/session files. This argument will be replaced by the corresponding image name.

    cache_dir: str
        The working directory of the workflow.

    output_dir: str
        The output folder path

    config_file: str
        Path to a configuration file. Will use anything compatible with Kaptan.
        Have a look at config.py.

    params: dict with arguments
        crumb_replaces

        atlas_file

        raise_on_filenotfound
    """
    attach_funcs, cfg_params, file_templates = _clinical_wf_setup(wf_name)

    if params is None:
        params = {}

    if config_file:
        update_config(config_file)

    if params:
        update_config(params)

    if cfg_params is not None:
        update_config(cfg_params)

    wf = build_crumb_workflow(attach_funcs,
                              data_crumb=data_crumb,
                              in_out_kwargs=file_templates,
                              output_dir=output_dir,
                              cache_dir=cache_dir,)

    return wf





# -*- coding: utf-8 -*-
"""
Helper functions for Slice Timing Correction
"""
import os.path as op

import nipype.pipeline.engine as pe
from   nipype.interfaces.base import traits, isdefined
from   nipype.interfaces.utility import IdentityInterface
from   nipype.algorithms.misc import Gunzip

from   .slicetime_params import STCParametersInterface
from   ..utils  import remove_ext
from   ..config import setup_node


def afni_slicetime(in_file=traits.Undefined,
                   out_file=traits.Undefined,
                   tr=traits.Undefined,
                   tr_units='s',
                   ignore_first=traits.Undefined,
                   slice_mode=traits.Undefined,
                   ref_slice=traits.Undefined,
                   tzero=traits.Undefined,
                   out_type='NIFTI_GZ'):
    """ Return a nipype interface to the AFNI 3dTshift command.

    Parameters
    ----------
    in_file: str
        Path to the input file

    out_file: str
        Path to the output file.

    tr: int or str
        The TR of the input dataset

    tr_units: str
        The units of `tr`. Choices: ('s', 'ms')

    ignore_first: int
        Number of acquisitions thrown away from the beginning of the
        dataset.

    tzero: float
        Align each slice to time offset `tzero`.
        The value of 'tzero' must be between the
        minimum and maximum slice temporal offsets.
        Note: The default alignment time is the average
              of the 'tpattern' values (either from the
              dataset header or from the -tpattern option)

    ref_slice: int
        Align each slice to the time offset of the given `ref_slice` index.
        Note: only one of the tzero and slice options can be used.
        -tslice flag

    slice_mode: str
        Choices:
        'alt+z' or 'altplus'    = alternating in the plus direction
        'alt+z2'                = alternating, starting at slice #1 instead of #0
        'alt-z' or 'altminus'   = alternating in the minus direction
        'alt-z2'                = alternating, starting at slice #nz-2 instead of #nz-1
        'seq+z' or 'seqplus'    = sequential in the plus direction
        'seq-z' or 'seqminus'   = sequential in the minus direction
         @filename              = read temporal offsets from 'filename'

    out_type: str
        ('NIFTI_GZ' or 'AFNI' or 'NIFTI')
        AFNI output filetype

    Returns
    -------
    tshift: nipype interface
    """
    import nipype.interfaces.afni as afni

    tshift = afni.TShift()

    if isdefined(tr):
        tr = '{}{}'.format(tr, tr_units)

    if isdefined(in_file):
        if not isdefined(out_file):
            out_file = op.join(op.dirname(in_file), 'tshift_' + remove_ext(op.basename(in_file)))

    tshift.inputs.in_file = in_file
    tshift.inputs.out_file = out_file
    tshift.inputs.tpattern = slice_mode
    tshift.inputs.tzero = tzero
    tshift.inputs.tr = tr
    tshift.inputs.tslice = ref_slice
    tshift.inputs.ignore = ignore_first
    tshift.inputs.outputtype = out_type

    #res = tshift.run()
    return tshift


def fsl_slicetime(in_file=traits.Undefined,
                  interleaved=True,):
    """
    Wraps command slicetimer
    use FSL slicetimer to perform slice timing correction.
    http://www.mit.edu/~satra/nipype-nightly/interfaces/generated/nipype.interfaces.fsl.preprocess.html#slicetimer

    Parameters
    ----------
    in_file
    interleaved

    Returns
    -------
    st: nipype Interface
    """
    from nipype.interfaces import fsl

    st = fsl.SliceTimer()
    st.inputs.in_file = in_file
    st.inputs.interleaved = interleaved
    return st


def spm_slicetime(in_files=traits.Undefined,
                  out_prefix=traits.Undefined,
                  num_slices=0,
                  time_repetition=-1,
                  time_acquisition=-1,
                  ref_slice=traits.Undefined,
                  slice_order=None,
                  ):
    """ Return a nipype interface to the SPM SliceTiming.

    Parameters
    ----------
    in_files: str or list of str
        Path to the input file(s).

    out_prefix: str
        Prefix to the output file.
        Default: 'a'

    num_slices: int

    time_repetition: int or str
        The time repetition (TR) of the input dataset in seconds
        Default: 0
        If left to default will read the TR from the nifti image header.

    time_acquisition: int
        Time of volume acquisition. usually calculated as TR-(TR/num_slices)

    ref_slice: int
        Index of the reference slice

    slice_order: list of int

    Returns
    -------
    stc: nipype interface
    """
    import nipype.interfaces.spm as spm

    stc = spm.SliceTiming()

    stc.inputs.in_files         = in_files
    stc.inputs.out_prefix       = 'a' if not isdefined(out_prefix) else out_prefix
    stc.inputs.slice_order      = slice_order
    stc.inputs.ref_slice        = ref_slice
    stc.inputs.time_repetition  = time_repetition
    stc.inputs.time_acquisition = time_acquisition
    stc.inputs.num_slices       = num_slices

    #res = stc.run()
    return stc


def nipy_fmrirealign4d(in_files=traits.Undefined,
                       time_repetition=2,
                       slice_order=None,
                       loops=5,
                       ):
    """ Return a nipype interface to the nipy.FmriRealign4d
    Simultaneous motion and slice timing correction algorithm.
    Parameters
    ----------
    in_files: str or list of str
        Path to the input file(s).

    time_repetition: int or str
        The time repetition (TR) of the input dataset in seconds
        Default: 0
        If left to default will read the TR from the nifti image header.

    slice_order: list of int
        This would be equivalent to
        entering np.argsort(spm_slice_order) for this field.
        This field will be deprecated in future Nipy releases and be
        replaced by actual slice acquisition
        times.

    loops: int
        Number of loops used to realignment runs.

    Returns
    -------
    stc: nipype interface
    """
    import nipype.interfaces.nipy as nipy

    stc = nipy.FmriRealign4d()

    stc.inputs.in_file     = in_files
    stc.inputs.tr          = time_repetition
    stc.inputs.slice_order = slice_order
    stc.inputs.loops       = loops

    #res = stc.run()
    return stc


def auto_spm_slicetime(in_file=traits.Undefined,
                       out_prefix='stc',
                       num_slices=traits.Undefined,
                       time_repetition=traits.Undefined,
                       time_acquisition=traits.Undefined,
                       ref_slice=traits.Undefined,
                       slice_order=traits.Undefined,
                       wf_name='auto_spm_slicetime'):
    """ A workflow that tries to automatically read the slice timing correction parameters
    from the input file and passes them to a spm.SliceTiming node.

    Parameters
    ----------
    in_files: str
        Path to the input file.

    out_prefix: str
        Prefix to the output file.
        Default: 'a'

    num_slices: int
        Number of slices of `in_files`.

    time_repetition: int or str
        The time repetition (TR) of the input dataset in seconds
        Default: 0
        If left to default will read the TR from the nifti image header.

    time_acquisition: int
        Time of volume acquisition. usually calculated as TR-(TR/num_slices)

    ref_slice: int
        Index of the reference slice

    slice_order: list of int
        List of integers with the order in which slices are acquired

    wf_name: str
        Name of the workflow

    Nipype Inputs
    -------------
    ## Mandatory:
    stc_input.in_files:

    ## Optional:
    stc_input.num_slices

    stc_input.slice_order

    stc_input.time_repetition

    stc_input.time_acquisition

    stc_input.ref_slice

    stc_input.slice_mode

    Nipype Outputs
    --------------
    stc_output.timecorrected_files

    stc_output.time_repetition

    Returns
    -------
    auto_spm_stc: nipype Workflow
        SPM slice timing correction workflow with automatic
        parameters detection.
    """
    # helper functions
    def _sum_one_to_each(slice_order): # SPM starts count from 1
        return [i+1 for i in slice_order]

    def _sum_one(num):
        return num + 1

    def _pick_first(sequence):
        return sequence[0]

    # the input and output nodes
    stc_input = setup_node(IdentityInterface(fields=["in_file",
                                                     "num_slices",
                                                     "slice_order",
                                                     "time_repetition",
                                                     "time_acquisition",
                                                     "ref_slice",
                                                     "slice_mode",
                                                    ]),
                                             name="stc_input")

    stc_output = setup_node(IdentityInterface(fields=["timecorrected_files",
                                                      "time_repetition",
                                                     ]),
                                               name="stc_output")

    # Declare the processing nodes
    params = setup_node(STCParametersInterface(in_files=in_file), name='stc_params')
    gunzip = setup_node(Gunzip(), name="gunzip")
    stc    = setup_node(spm_slicetime(out_prefix       = out_prefix,
                                      num_slices       = num_slices,
                                      time_repetition  = time_repetition,
                                      time_acquisition = time_acquisition,
                                      ref_slice        = ref_slice,
                                      slice_order      = slice_order), name='slice_timer')

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # Connect the nodes
    wf.connect([
                # input node
                (stc_input, params, [("in_file",          "in_files"),
                                     ("num_slices",       "num_slices"),
                                     ("slice_order",      "slice_order"),
                                     ("time_repetition",  "time_repetition"),
                                     ("time_acquisition", "time_acquisition"),
                                     ("ref_slice",        "ref_slice"),
                                     ("slice_mode",       "slice_mode"),
                                    ]),

                # processing nodes
                (params, gunzip,    [(("in_files",    _pick_first),      "in_file")]),
                (params, stc,       [(("slice_order", _sum_one_to_each), "slice_order"),
                                     (("ref_slice",   _sum_one),         "ref_slice"),
                                     ("num_slices",                      "num_slices"),
                                     ("time_acquisition",                "time_acquisition"),
                                     ("time_repetition",                 "time_repetition"),
                                    ]),
                (gunzip, stc,       [("out_file",                        "in_files")]),

                # output node
                (params, stc_output,[("time_repetition",     "time_repetition")]),
                (stc,    stc_output,[("timecorrected_files", "timecorrected_files")]),
              ])

    return wf


def auto_nipy_slicetime(in_files=traits.Undefined,
                        time_repetition=traits.Undefined,
                        slice_order=traits.Undefined,
                        loops=5,
                        wf_name='auto_nipy_slicetime'):
    """ A workflow that tries to automatically read the slice timing correction parameters
    from the input files and passes them to a nipy.fMRIRealign4D node.

    Parameters
    ----------
    in_file: str or list of str
        Path to the input file(s).

    time_repetition: int or str
        The time repetition (TR) of the input dataset in seconds
        Default: 0
        If left to default will read the TR from the nifti image header.

    slice_order: list of int
        List of integers with the order in which slices are acquired

    loops: int
        Number of loops used to realignment runs.

    wf_name: str
        Name of the workflow

    Nipype Inputs
    -------------
    ## Mandatory:
    params.in_files:

    params.time_repetition

    ## Optional:
    params.slice_order

    params.ref_slice

    params.slice_mode

    Nipype Outputs
    --------------
    slice_timer.out_file

    slice_timer.par_file

    Returns
    -------
    auto_nipy_stc: nipype Workflow
        Nipy 4D alignment and slice timing correction workflow with automatic
        parameters detection.
    """
    # Declare the processing nodes
    params = setup_node(STCParametersInterface(in_files=in_files), name='stc_params')
    stc    = setup_node(nipy_fmrirealign4d(time_repetition=time_repetition,
                                           slice_order=slice_order,
                                           loops=loops),
                        name='slice_timer')

    # Create the workflow object
    wf = pe.Workflow(name=wf_name)

    # Connect the nodes
    wf.connect([
                (params, stc, [("in_files",         "in_files"),
                               ("slice_order",      "slice_order"),
                               ("time_repetition",  "tr"),
                              ]),
              ])

    return wf